#pragma once

#include "ofMain.h"

//uuid
#include "ofRectangle.h"
#include "random_id.h"

#include "ofxMsdfgen.h"
#include "Utils.h"
#include <memory>
#include <sstream>

namespace ofxVariableLab {

struct LayerSettings {
    VFlipBehaviour vFlipBehaviour = V_FLIP_ONCE_AUTO;
    float letterDelayRatio = 30.0f; //
};

class Layer {
    public:
        enum TransformOrigin : int {
            TOP_LEFT = 0,
            TOP_RIGHT,
            CENTER,
            BOTTOM_LEFT,
            BOTTOM_RIGHT
        };
        struct Props {
            float x = 200;
            float y = 200;
            float width = 0;
            float height = 0;
            float rotation = 0;
            float fontSize_px = 42;
            float letterSpacing = 0;
            float lineHeight = 1;
            float textAlignment = 0;
            std::array <float, 4> color = {0, 0, 0, 1};
            bool mirror_x = false;
            float mirror_x_distance = 0;
            bool mirror_y = false;
            float mirror_y_distance = 0;
            bool mirror_xy = false;
            float letterDelay = 0;
            int transformOrigin = TransformOrigin::TOP_LEFT;
            std::vector <ofxVariableLab::FontVariation> fontVariations;
            std::map <string, float> letterDelays;
            string text = "abc";
            string fontPath = "data/celines-fonts/Version-2-var.ttf";
        };
        struct BoundingBox {
            float x;
            float y;
            float w;
            float h;
        };

        static std::string asString(const ofxVariableLab::Layer::Props & props){
            stringstream s;
            s << "asString::props:" << endl
              << "\tx:" << ofToString(props.x) << endl
              << "\ty:" << ofToString(props.y) << endl
              << "\trotation:" << ofToString(props.rotation) << endl
              << "\tfontSize_px:" << ofToString(props.fontSize_px) << endl
              << "\tcolor: ["
              << ofToString(props.color[0]) << ","
              << ofToString(props.color[1]) << ","
              << ofToString(props.color[2]) << ","
              << ofToString(props.color[3]) << "]"
              << endl
              << "\tmirror_x:" << ofToString(props.mirror_x) << endl
              << "\tmirror_x_distance:" << ofToString(props.mirror_x_distance) << endl
              << "\tmirror_y:" << ofToString(props.mirror_y) << endl
              << "\tmirror_y_distance:" << ofToString(props.mirror_y_distance) << endl
              << "\tletterDelay:" << ofToString(props.letterDelay) << endl
              << "\ttransformOrigin:" << ofToString(props.transformOrigin) << endl;
            for(const auto & fv : props.fontVariations){
                s << "\tfontVariations (" << fv.name << "):" << ofToString(fv.value) << endl;
            }
            s << "\ttext: " << props.text << endl
              << "\tfontPath: " << props.fontPath << endl
            ;
            return s.str();
        }

        virtual void setup(const LayerSettings & settings = LayerSettings()) = 0;
        virtual void update() = 0;
        virtual void draw(glm::vec3 position = glm::vec3(0, 0, 0)) = 0;
        virtual void drawCharacter(const char character,
                                   glm::vec3 position = glm::vec3(0, 0, 0),
                                   glm::vec4 color = glm::vec4(1, 1, 1, 1),
                                   float rotation = 0,
                                   float scale = 1,
                                   ofxVariableLab::FontVariation fontVariation = ofxVariableLab::FontVariation()) = 0;
        virtual void setProps(const Props & props) = 0;
        virtual const Props & getProps(float delay = 0) = 0;
        virtual void clearPropsBuffer() = 0;
        virtual void setId(const LayerID & id) = 0;
        virtual const LayerID & getId() = 0;
        virtual void setShader(shared_ptr <ofShader> _shader) = 0;
        virtual shared_ptr <ofShader> getShader() const = 0;
        virtual const LayerType & getType() const = 0;
        virtual void setVFlip(const VFlipState vFlip) = 0;
        virtual bool isDirtyDirty() const = 0;
        virtual void setDirtyDirty(bool dirtyDirty) = 0;
        virtual bool wantsNewMom() = 0;
        virtual void isWithNewMom() = 0;
        virtual const ComboIdentifier & getMomsComboIdentifier() const = 0;
        virtual void setMomsComboIdentifier(const ComboIdentifier & identifier) = 0;
        virtual ofNode & getOuterNode() = 0;
        virtual ofNode & getInnerNode() = 0;
        virtual const Layer::BoundingBox & getBoundingBox() = 0;
        virtual void setBoundingBox(const Layer::BoundingBox & boundingBox) = 0;
        static int n_layers;
};
}
